/*

  qh-math.c -- MathLink(TM) interface to Qhull's Delaunay Triangulation  

  Original work Copyright (c) 1998 by Alban P.M. Tsui <a.tsui@ic.ac.uk> 
  Modifications Copyright (c) 2000 by P.J. Hinton <paulh@wolfram.com>
 
  Author: Alban P.M. Tsui
  Date:   Jan 19, 1998
  Tested: Unix, Windows95

  Revised by P.J. Hinton, July 5, 2000
  Ported: Linux (glibc 2.1), Windows 95, MacOS System 8.6
  
  Use of this source code implies consent to the license in the file
  COPYING that accompanies this source code archive.

*/

#include <stdio.h>
#include <stdlib.h>

#include "mathlink.h"
#include "qhull.h"
#include "mem.h"
#include "qset.h"
#include "qhull_a.h"

#define BUF_SIZE 2048
#define QHULL_COMMAND "qhull d QJ Tv "

/* Function Prototypes */

void delaunayQhull(void);
void kernel_message(char* symbol, char* tag);

char qh_version[] = "qh-math.c 2000/7/6";

void delaunayQhull(void) {

	int curlong, totlong, exitcode, dim;
	long temp, numpoints, numfacets, useroptslen;
	double *a;
	char options [BUF_SIZE];
	const unsigned char *useropts;
	int i,j,k;
	coordT *point, *points;

	facetT *facet;
	vertexT *vertex, **vertexp;

	/* Verify that expression waiting to be read from link has 
		the head List.  Get number of elements in list */

	MLCheckFunction(stdlink, "List", &numpoints);

	/* Pull first element off the link, assuming that it is a 
		list of Reals.  Assume that length of list determines
		the dimension of the coordinate space. */

	if ( MLGetRealList(stdlink, &a, &temp) == 0 ) {
		MLClearError(stdlink);
		MLPutFunction(stdlink, "CompoundExpression", 2);
			/* Message[qDelaunayTriangulation::"notreallist"] */
				kernel_message("qDelaunayTriangulation", "notreallist");
				MLPutSymbol(stdlink, "$Failed");
		return;
	}

	dim = temp;

	/* Allocate memory to store elements of the list */

	points = (coordT*) malloc((dim+1)*numpoints*sizeof(coordT));

	/* Store each coordinate in an array element. */

	for(i=0; i<dim; i++) {
		points[i] = a[i];
	}

	MLDisownRealList(stdlink, a, temp);

	/* Get the rest of the elements of the list. */

	for (j=1; j < numpoints; j++) {

		point = points + j*(dim+1);

		/* Make sure that get operation succeeds */
		if ( MLGetRealList(stdlink, &a, &temp) == 0 ) {
			MLClearError(stdlink);
			MLPutFunction(stdlink, "CompoundExpression", 2);
				/* Message[qDelaunayTriangulation::"notreallist"] */
				kernel_message("qDelaunayTriangulation", "notreallist");
				MLPutSymbol(stdlink, "$Failed");
			return;	
		}
		
		/* Make sure that the dimensions of the current list match that
			of the first list */
	
		if (temp != dim) {
			MLClearError(stdlink);
			MLPutFunction(stdlink, "CompoundExpression", 2);
				/* Message[qDelaunayTriangulation::"unequivdim"] */
				kernel_message("qDelaunayTriangulation", "unequivdim"); 
				MLPutSymbol(stdlink, "$Failed");
				MLDisownRealList(stdlink, a, temp);
			return;
		}

		for (k= 0; k < dim; k++) {
			point[k] = a[k];
 		}

 		MLDisownRealList(stdlink, a, temp);
	}

	/* Retrieve command line options supplied by user. */

	MLGetByteString(stdlink, &useropts, &useroptslen, 0L);
 
	/* Call out to qhull library to compute DeLaunay triangulation. */

	qh_init_A(stdin, stdout, stderr, 0, NULL);
	exitcode= setjmp(qh errexit);
	
	if (!exitcode) {
	
		/* Use options only if we're not in danger of buffer overflow. */
		if ( (useroptslen + strlen(QHULL_COMMAND) + 1 ) <= BUF_SIZE ) {
			sprintf(options,  QHULL_COMMAND, useropts);
		} else {
			strcpy(options, QHULL_COMMAND);
		}
	
		/* User options no longer needed, so the memory allocated to it is freed */
	
		MLDisownByteString(stdlink, useropts, useroptslen);
	
		qh_initflags(options);
 		qh_setdelaunay(dim+1, numpoints, points);
		qh_init_B(points, numpoints, dim+1, False);
		qh_qhull();
		qh_check_output();

		/* Send triangulation across link back to kernel. */

		numfacets = 0;

		FORALLfacets {
			if (!facet->upperdelaunay){
				numfacets+=1;
			}
		}

		MLPutFunction(stdlink, "List", numfacets);

		FORALLfacets {
			if (!facet->upperdelaunay){
				MLPutFunction(stdlink, "List", dim+1);
				FOREACHvertex_(facet->vertices){
					MLPutInteger(stdlink, qh_pointid(vertex->point)+1);
				}
			}
		}

	}

	/* Free allocated memory. */

	qh NOerrexit= True;
	qh_freeqhull (False);
	qh_memfreeshort (&curlong, &totlong);
	free(points);
	MLFlush(stdlink);
	
	if (curlong || totlong)
		fprintf (stderr, "qhull internal warning (main): did not \
			free %d bytes of long memory (%d pieces)\n", totlong, curlong);

	return;

}

#ifdef WINDOWS_MATHLINK

#if __BORLANDC__
#pragma argsused
#endif

int PASCAL WinMain( HINSTANCE hinstCurrent, HINSTANCE hinstPrevious, LPSTR lpszCmdLine, int nCmdShow)
{
	char  buff[512];
	char FAR * buff_start = buff;
	char FAR * argv[32];
	char FAR * FAR * argv_end = argv + 32;

	hinstPrevious = hinstPrevious; /* suppress warning */

	if( !MLInitializeIcon( hinstCurrent, nCmdShow)) return 1;
	MLScanString( argv, &argv_end, &lpszCmdLine, &buff_start);
	return MLMain( argv_end - argv, argv);
}

#elif MACINTOSH_MATHLINK

int main( int argc, char* argv[])
{
        /* Due to a bug in some standard C libraries that have shipped with
         * MPW, zero is passed to MLMain below.  (If you build this program
         * as an MPW tool, you can change the zero to argc.)
         */
        argc = argc; /* suppress warning */
        return MLMain( 0, argv);
}

#else /* UNIX */

int main(int argc, char *argv[])
{
	return MLMain(argc, argv);
}

#endif

void kernel_message(char* symbol, char* tag) 
{

	/* Message[symbol::"tag"] */
	MLPutFunction(stdlink, "Message", 1);
		MLPutFunction(stdlink, "MessageName", 2);
			MLPutSymbol(stdlink, symbol); 
			MLPutString(stdlink, tag);

	return;

}
